@@ -7,7 +7,7 @@ from e3nn_jax._src.util.prod import prod
 
 
 def test_empty():
-    x = e3nn.IrrepsArray.from_list("", [], (2, 2))
+    x = e3nn.IrrepsArray.from_list("", [], (2, 2), jnp.float32)
     assert x.irreps == e3nn.Irreps([])
     assert x.shape == (2, 2, 0)
 
@@ -105,6 +105,8 @@ def test_operators():
     with pytest.raises(ValueError):
         x * jnp.array([1.0, 2.0, 3.0, 4.0, 5.0])
 
+    x * jnp.array([1.0, 2.0, 3.0])
+
     with pytest.raises(ValueError):
         1.0 / x
 
